import optuna
import pandas as pd
from sklearn.model_selection import KFold
import statistics as st
import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader

from Constants import Constants
from GeneralUtils import group_and_pad_sensor_data, scale_features, get_dev_training_data, DatasetDirectoryNames, \
    set_deterministic_torch, PatientDataset
from GRU_Dense_NN import GruDenseNn, prepare_and_convert_to_tensor
from functools import partial

kfold_seed = 1  # The kfold seed is only useful if kfold_shuffle is True
kfold_shuffle = False

should_scale_features = True
did_drop_first_last_trend_from_df = True

if did_drop_first_last_trend_from_df:
    data_folder_suffix_from_dropped = "_DROP_FIRST_LAST_TREND"
else:
    data_folder_suffix_from_dropped = ""

data_directory = DatasetDirectoryNames.firstLastNans
tunedHyperparamsFile = f"best_params_{data_directory}{data_folder_suffix_from_dropped}.csv"
set_deterministic_torch(Constants.torch_seed)


def objective(trial, X_static, X_sensor, y, k_folds):
    print(f"Starting trial: {trial.number}".format(trial))

    trial_lr = trial.suggest_float('learning_rate', 0.01, 0.1, step=0.01)
    trial_gru_hidden_size = trial.suggest_int('hidden_size', 16, 16, step=8)
    trial_n_epochs = trial.suggest_int('n_epochs', 1, 301, step=5)
    trial_gru_n_layers = trial.suggest_int('n_layers', 1, 3, step=1)
    trial_dropout = trial.suggest_float('dropout', 0.1, 0.5, step=0.02)
    trial_batch_size_divider = trial.suggest_int('batch_size_divider', 2, 6, step=2)

    if trial_gru_n_layers <= 1:
        trial_dropout = 0.0  # dropout does not apply to last layer anyway

    patient_ids = X_static['patient_id'].unique()

    # cross validation
    kf = KFold(n_splits=k_folds)
    cross_validation_losses = []

    for train_index, validation_index in kf.split(patient_ids):
        train_ids, validation_ids = patient_ids[train_index], patient_ids[validation_index]

        # TODO: Split the tensor input parameters based on the train_ids and validation_ids
        # Split the tensor input parameters based on the train_ids and validation_ids
        X_train_static = X_static[X_static['patient_id'].isin(train_ids)]
        X_val_static = X_static[X_static['patient_id'].isin(validation_ids)]

        X_train_sensor = X_sensor[X_sensor['patient_id'].isin(train_ids)]
        X_val_sensor = X_sensor[X_sensor['patient_id'].isin(validation_ids)]

        y_train = y[y['patient_id'].isin(train_ids)]
        y_val = y[y['patient_id'].isin(validation_ids)]

        if should_scale_features:
            scaled_train_sets, scaled_test_sets = scale_features([X_train_static, X_train_sensor],
                                                                 [X_val_static, X_val_sensor])
            X_train_static = scaled_train_sets[0]
            X_train_sensor = scaled_train_sets[1]

            X_val_static = scaled_test_sets[0]
            X_val_sensor = scaled_test_sets[1]

        X_train_static_tensor = prepare_and_convert_to_tensor(X_train_static)
        y_train_tensor = prepare_and_convert_to_tensor(y_train)

        X_val_static_tensor = prepare_and_convert_to_tensor(X_val_static)
        y_val_tensor = prepare_and_convert_to_tensor(y_val)

        X_train_sensor_tensor = group_and_pad_sensor_data(X_train_static, X_train_sensor)
        X_val_sensor_tensor = group_and_pad_sensor_data(X_val_static, X_val_sensor)

        train_dataset = PatientDataset(X_train_static_tensor, X_train_sensor_tensor, y_train_tensor)
        # val_dataset = PatientDataset(X_val_static_tensor, X_val_sensor_tensor, y_val_tensor)

        max_batch_size = X_train_static_tensor.shape[0]

        batch_size = max_batch_size // trial_batch_size_divider
        train_dataloader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

        n_sensor_features = X_train_sensor_tensor.shape[2]
        n_static_features = X_train_static_tensor.shape[1]
        output_size = y_train_tensor.shape[1]

        compressed_data_size = n_sensor_features + n_static_features

        trial_fc_hidden_size = trial.suggest_int('trial_fc_layer_size', output_size, compressed_data_size, step=2)

        model = GruDenseNn(gru_input_size=n_sensor_features,
                           gru_hidden_size=trial_gru_hidden_size,
                           gru_num_layers=trial_gru_n_layers,
                           static_data_size=n_static_features,
                           ffnn_hidden_size=trial_fc_hidden_size,
                           ffnn_output_size=output_size,
                           gru_dropout=trial_dropout)

        optimizer = torch.optim.Adam(model.parameters(), lr=trial_lr)

        # Training
        for epoch in range(trial_n_epochs):
            batchLosses = []
            n_batches = 0
            for X_static, X_sensor, y in train_dataloader:
                n_batches += 1
                # Clear the gradients
                optimizer.zero_grad()

                # Forward pass
                output = model(X_sensor, X_static)

                # Compute the loss
                loss = model.mean_mae_rmse_loss(output, y)
                loss.backward()

                optimizer.step()
                batchLosses.append(loss.item())

            if (epoch + 1) % 10 == 0:
                print(f'Epoch {epoch + 1}/{trial_n_epochs}, Loss: {st.mean(batchLosses)}')

        # Validation
        model.eval()
        with torch.no_grad():
            predictions = model(X_val_sensor_tensor, X_val_static_tensor)
            predictions = torch.round(predictions)
            predictions = torch.clamp(predictions,
                                      min=Constants.min_prediction,
                                      max=Constants.max_prediction)

            validation_loss = model.mean_mae_rmse_loss(predictions, y_val_tensor)
            cross_validation_losses.append(validation_loss.item())

    mean_val_loss = st.mean(cross_validation_losses)
    print(f'Trial {trial.number} Mean Validation Loss: {mean_val_loss} \n')

    return mean_val_loss


if __name__ == "__main__":

    X_train_static, X_train_sensor, y_train = get_dev_training_data(
        f"{data_directory}{data_folder_suffix_from_dropped}",
        include_split_sensor=True)

    study = optuna.create_study(direction='minimize')
    try:
        study.optimize(partial(objective,
                               X_static=X_train_static,
                               X_sensor=X_train_sensor,
                               y=y_train,
                               k_folds=2), n_trials=6)
    except KeyboardInterrupt:
        print("\nOptimization interrupted by user.")
    finally:
        print("Study statistics")
        trial = study.best_trial
        best_params = study.best_params

        df = pd.DataFrame.from_dict(best_params, orient='index').transpose()
        df.to_csv(f'best_params_{tunedHyperparamsFile}.csv', index=False)

        print("Best trial mean cross-val MAE: ", trial.value)
        print("Best trial params: ")

        for param in best_params:
            print(f"{param}: {best_params[param]}")

            # Save data from all trials to a CSV file
            df_all_trials = study.trials_dataframe()
            df_all_trials.to_csv(f'all_trials_{tunedHyperparamsFile}.csv', index=False)
