import torch.nn as nn
import numpy as np
from GRU import SimpleGRU
import torch


def prepare_and_convert_to_tensor(df):
    # Remove the 'patient_id' column
    df = df.drop(columns='patient_id')

    # Convert the dataframe to a PyTorch tensor
    tensor = torch.tensor(df.values, dtype=torch.float32)

    return tensor


def prepare_sensor_data_for_gru(X_sensor):
    # Group by patient_id and count the number of entries for each patient
    entries_per_patient = X_sensor.groupby('patient_id').size()

    # Find the maximum number of entries
    num_days = entries_per_patient.max()

    # Create a list to store the sequences
    sensor_data_list = []

    # Group the sensor data by patient_id
    grouped = X_sensor.groupby('patient_id')

    # For each patient
    for patient_id, group in grouped:
        # Get the sensor data for the patient and convert it to a numpy array
        sensor_data_np = group.drop(columns='patient_id').to_numpy()

        # If the patient has less than num_days entries
        if sensor_data_np.shape[0] < num_days:
            # Pad the sequence with zeros
            sensor_data_np = np.pad(sensor_data_np, ((0, num_days - sensor_data_np.shape[0]), (0, 0)), 'constant',
                                    constant_values=0)

        # Add the sequence to the list
        sensor_data_list.append(sensor_data_np)

    # Convert the list to a numpy array
    sensor_data_np = np.array(sensor_data_list)

    return sensor_data_np


class GruDenseNn(nn.Module):
    def __init__(self,
                 gru_input_size,
                 gru_hidden_size,
                 gru_num_layers,
                 static_data_size,
                 ffnn_hidden_size,
                 ffnn_output_size,
                 gru_dropout=0.0):

        super(GruDenseNn, self).__init__()
        # GRU for compressing the sensor data
        self.gru_num_layers = gru_num_layers
        self.gru_hidden_size = gru_hidden_size

        self.gru = SimpleGRU(n_layers=gru_num_layers,
                             input_size=gru_input_size,
                             hidden_size=gru_hidden_size,
                             dropout=gru_dropout)

        # Should have the same variable that sets the output_size of the GRU and the input_size of the FFNN
        # Feed-Forward Neural Network for prediction
        self.ffnn = nn.Sequential(
            nn.Linear(gru_hidden_size + static_data_size, ffnn_hidden_size),
            nn.ReLU(),
            nn.Linear(ffnn_hidden_size, ffnn_output_size)
        )

    def forward(self, sensor_data, static_data):
        # Pass sensor data through GRU
        # Important: Even though multiple entries might correspond to the same patient_id in training, they should be
        # treated as individual entries and the hidden layer(s) should probably be set to zeros after each batch of sensor
        # data processed (from init_t to target_t). This (should) help because the testing/validation sets have only one
        # entry/row per patient_id and thus it would be? pointless to learn temporal relations between multiple entries
        # on the train set.
        # So the GRU should only use its memory between sensor data days in the same batch
        # (corresponding to a single entry) and will reset hidden layer(s) afterward. This behaviour should be checked.
        # h0 = torch.zeros(self.gru_num_layers, sensor_data.size(0), self.gru_hidden_size).to(sensor_data.device)

        compressed_sensor_data = self.gru(sensor_data)
        #print(compressed_sensor_data.shape)

        # Use the last hidden state from the GRU as the compressed sensor data
        compressed_sensor_data = compressed_sensor_data[:, -1, :]

        # print("Shape of compressed_sensor_data:", compressed_sensor_data.shape)
        # print("Shape of static data:", static_data.shape)
        # Concatenate the compressed sensor data with the static data
        combined_data = torch.cat((compressed_sensor_data, static_data), dim=1)

        # Pass the combined data through the FFNN
        output = self.ffnn(combined_data)

        return output

    def mae_loss(self, y_pred, y_true):
        return torch.mean(torch.abs(y_pred - y_true))

    def rmse_loss(self, y_pred, y_true):
        mse = torch.mean((y_pred - y_true) ** 2)
        return torch.sqrt(mse)

    def mean_mae_rmse_loss(self, y_pred, y_true):
        mae_loss = self.mae_loss(y_pred, y_true)
        rmse_loss = self.rmse_loss(y_pred, y_true)
        return (mae_loss + rmse_loss) / 2
