import pandas as pd
import torch
from torch.utils.data import DataLoader
import statistics as st

from GeneralUtils import set_deterministic_torch, scale_features, group_and_pad_sensor_data, PatientDataset, \
    generate_submission, fillNansBasic
from GRU_Dense_NN import GruDenseNn, prepare_and_convert_to_tensor
from Constants import Constants

test_sensor = pd.read_csv('Data/test-sensor.csv')
test_sensor_interpolated = pd.read_csv('Data/test-sensor_interpolated.csv')
test_sensor_interpolated_strip_87 = pd.read_csv('Data/test-sensor_interpolated_strip87.csv')
test_static = pd.read_csv('Data/test-static.csv')
train_sensor = pd.read_csv('Data/train-sensor.csv')
train_sensor_interpolated = pd.read_csv('Data/train-sensor_interpolated.csv')
train_sensor_interpolated_strip_87 = pd.read_csv('Data/train-sensor_interpolated_strip87.csv')
train_static = pd.read_csv('Data/train-static.csv')
val = pd.read_csv('Data/val.csv')

# print("Test sensor shape", test_sensor.shape)
# print("Test sensor interpolated shape", test_sensor_interpolated.shape)
# print("Test sensor interpolated strip 87 shape", test_sensor_interpolated_strip_87.shape)
# print("Test static shape", test_static.shape)
# print("Train sensor shape", train_sensor.shape)
# print("Train sensor interpolated shape", train_sensor_interpolated.shape)
# print("Train sensor interpolated strip 87 shape", train_sensor_interpolated_strip_87.shape)
# print("Train static shape", train_static.shape)
# print("Val shape", val.shape)


train_sensor_interpolated_strip_87 = train_sensor_interpolated_strip_87.drop(columns=['days_from_start'])
test_sensor_interpolated_strip_87 = test_sensor_interpolated_strip_87.drop(columns=['days_from_start'])
train_sensor_interpolated = train_sensor_interpolated.drop(columns=['days_from_start'])
test_sensor_interpolated = test_sensor_interpolated.drop(columns=['days_from_start'])

current_train_sensor = fillNansBasic(train_sensor_interpolated)
current_test_sensor = fillNansBasic(test_sensor_interpolated)
current_train_static = fillNansBasic(train_static)
current_test_static = fillNansBasic(test_static)

y_train = val

# Configuration
set_deterministic_torch(Constants.torch_seed)
tuned_hyperparams_file = ""

if __name__ == "__main__":
    # Preprocessing
    scaled_train_sets, scaled_test_sets = scale_features([current_train_static, current_train_sensor],
                                                         [current_test_static, current_test_sensor])

    X_train_static = scaled_train_sets[0]
    X_train_sensor = scaled_train_sets[1]

    X_test_static = scaled_test_sets[0]
    X_test_sensor = scaled_test_sets[1]

    X_train_static_tensor = prepare_and_convert_to_tensor(X_train_static)
    y_train_tensor = prepare_and_convert_to_tensor(y_train)

    X_test_static_tensor = prepare_and_convert_to_tensor(X_test_static)

    X_train_sensor_tensor = group_and_pad_sensor_data(X_train_static, X_train_sensor)
    X_test_sensor_tensor = group_and_pad_sensor_data(X_test_static, X_test_sensor)

    print("X_train_sensor shape:", X_train_sensor_tensor.shape)
    print("X_test_sensor shape:", X_test_sensor_tensor.shape)

    print("X_train_static shape", X_train_static_tensor.shape)
    print("X_test_static shape", X_test_static_tensor.shape)

    print("y_train shape", y_train_tensor.shape)

    # Create data loader to train in batches
    train_dataset = PatientDataset(X_train_static_tensor, X_train_sensor_tensor, y_train_tensor)

    max_batch_size = X_train_static_tensor.shape[0]

    batch_size = max_batch_size // Constants.batch_size_divider  # You can adjust this value
    train_dataloader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

    # Configure the model
    n_sensor_features = X_train_sensor_tensor.shape[2]
    n_static_features = X_train_static_tensor.shape[1]
    output_size = y_train_tensor.shape[1]

    # Define the model and optimizer
    model = GruDenseNn(gru_input_size=n_sensor_features,
                       gru_hidden_size=Constants.gru_hidden_size,
                       gru_num_layers=Constants.gru_n_layers,
                       static_data_size=n_static_features,
                       ffnn_hidden_size=Constants.ffnn_hidden_size,
                       ffnn_output_size=output_size,
                       gru_dropout=Constants.dropout)

    optimizer = torch.optim.Adam(model.parameters(), lr=Constants.learning_rate)

    # Training
    n_epochs = Constants.n_epochs
    for epoch in range(n_epochs):
        batchLosses = []
        for X_static, X_sensor, y in train_dataloader:
            # Clear the gradients
            optimizer.zero_grad()

            # Forward pass
            output = model(X_sensor, X_static)

            # Compute the loss
            loss = model.mean_mae_rmse_loss(output, y)
            loss.backward()

            optimizer.step()
            batchLosses.append(loss.item())

        if (epoch + 1) % 5 == 0:
            print(f'Epoch {epoch + 1}/{n_epochs}, Loss: {st.mean(batchLosses)}')

    # Generate predictions on test set
    model.eval()
    with torch.no_grad():
        predictions = model(X_test_sensor_tensor, X_test_static_tensor)
        predictions = torch.round(predictions)
        predictions = torch.clamp(predictions,
                                  min=Constants.min_prediction,
                                  max=Constants.max_prediction)

    # TODO: Print hyperparameter values
    print("Batch size divider:", Constants.batch_size_divider)
    print("Num epochs:", n_epochs)
    print("Gru hidden size:", Constants.gru_hidden_size)
    print("Gru num_layers:", Constants.gru_n_layers)
    print("Ffnn hidden size:", Constants.ffnn_hidden_size)
    print("Learning rate:", Constants.learning_rate)
    print("Dropout:", Constants.dropout)

    print("Predictions: ", predictions)

    import matplotlib.pyplot as plt
    import seaborn as sns

    df_pred = pd.DataFrame(predictions)
    df_pred.columns = [f'i{j}' for j in
                       ['Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6', 'Q7', 'Q8', 'Q9', 'Q10', 'Q11', 'Q12']]
    print(df_pred)
    df_init = current_test_static[[f'i{j}' for j in
                             ['Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6', 'Q7', 'Q8', 'Q9', 'Q10', 'Q11', 'Q12']]]
    print(df_init)
    diff = df_pred - df_init
    print(diff)
    plt.figure(figsize=(10, 10))
    sns.heatmap(diff, cmap='coolwarm', annot=True, linewidths=.5)

    plt.show()

    generate_submission(X_test_static, df_pred, 'GRU3')
