import torch
from sklearn.preprocessing import MinMaxScaler
from torch.nn.utils.rnn import pad_sequence
import pandas as pd
from LSTMUtils import Constants
from LSTMUtils import SimpleLSTM
from GeneralUtils import create_tensor_from, plot_loss_per_question, scale_features

should_normalize_features = True

if __name__ == "__main__":
    torch.manual_seed(Constants.torch_seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    firstLastNans = "first_last_nans"
    interpolatedSPODropped = "first_last_nans_interpolated_spo_dropped"
    firstLastNansInterpolatedTrend5 = "first_last_nans_interpolated_trend5_strip87"
    firstLastNansInterpolatedTrend3 = "first_last_nans_interpolated_trend3_strip87"

    dev_data_folder = firstLastNansInterpolatedTrend5
    X_test = pd.read_csv(f'../../dev_data/{dev_data_folder}/X_test.csv')
    y_test = pd.read_csv(f'../../dev_data/{dev_data_folder}/Y_test.csv')
    X_train = pd.read_csv(f'../../dev_data/{dev_data_folder}/X_train.csv')
    y_train = pd.read_csv(f'../../dev_data/{dev_data_folder}/Y_train.csv')

    if should_normalize_features:
        X_train, X_test = scale_features(X_train,
                                         X_test,
                                         scaler=MinMaxScaler(feature_range=(-1, 1)))

    useTunedHyperparameters = False
    tunedHyperparamsFile = "best_params.csv"

    if useTunedHyperparameters:
        print(f"Using AutoTuned Hyperparameters from {tunedHyperparamsFile}.")
        Constants.apply_best_param_config(tunedHyperparamsFile)
    else:
        print(f"Not using autotuned hyperparameters.")

    # Convert data to lists of torch tensors (prepare for training)
    X_train_tensors = [create_tensor_from(group) for _, group in X_train.groupby('patient_id')]
    y_train_tensors = [create_tensor_from(group) for _, group in y_train.groupby('patient_id')]
    X_test_tensors = [create_tensor_from(group) for _, group in X_test.groupby('patient_id')]
    y_test_tensors = [create_tensor_from(group) for _, group in y_test.groupby('patient_id')]

    # Pad the sequences
    X_train = pad_sequence(X_train_tensors, batch_first=True)
    y_train = pad_sequence(y_train_tensors, batch_first=True)

    X_test = pad_sequence(X_test_tensors, batch_first=True)
    y_test = pad_sequence(y_test_tensors, batch_first=True)

    n_features = X_train.shape[-1]
    prediction_length = y_train.shape[-1]

    model = SimpleLSTM(n_layers=Constants.n_layers,
                       hidden_size=Constants.hidden_size,
                       input_size=n_features,
                       output_size=prediction_length)

    optimizer = torch.optim.Adam(model.parameters(), lr=Constants.learning_rate)

    # Training
    for epoch in range(Constants.n_epochs):
        model.zero_grad()
        output = model(X_train)

        mae_loss = model.mae_loss(output, y_train)
        rmse_loss = model.rmse_loss(output, y_train)

        loss = (mae_loss + rmse_loss) / 2
        loss.backward()

        # Gradient clipping
        # torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1)

        optimizer.step()
        if epoch % 10 == 0:
            print(f'Epoch: {epoch}, Avg(MAE, RMSE) Loss: {loss.item()}')

        # Evaluation
    model.eval()
    with torch.no_grad():
        predictions = model(X_test)
        predictions = torch.round(predictions)
        predictions = torch.clamp(predictions,
                                  min=Constants.min_prediction,
                                  max=Constants.max_prediction)

        mae_test_loss = model.mae_loss(predictions, y_test)
        rmse_test_loss = model.rmse_loss(predictions, y_test)

        print(f'\nTest Loss Avg(MAE, RMSE): {(mae_test_loss + rmse_test_loss) / 2}\n')

        print(f'\nMAE Test Loss: {mae_test_loss.item()}')
        print(f'RMSE Test Loss: {rmse_test_loss.item()} \n')

        # Initialize lists to store MAE and RMSE for each question
        mae_test_loss_per_question = []
        rmse_test_loss_per_question = []
        total_loss_per_question = []

        # Loop over the third dimension of the tensors (i.e., the questions)
        for i in range(predictions.shape[2]):
            mae_test_loss = model.mae_loss(predictions[:, :, i], y_test[:, :, i])
            rmse_test_loss = model.rmse_loss(predictions[:, :, i], y_test[:, :, i])
            total_loss = (mae_test_loss.item() + rmse_test_loss.item()) / 2

            mae_test_loss_per_question.append(mae_test_loss.item())
            rmse_test_loss_per_question.append(rmse_test_loss.item())
            total_loss_per_question.append(total_loss)

        # plot_loss_per_question(total_losses=total_loss_per_question,
        #                        mae_losses=mae_test_loss_per_question,
        #                        rmse_losses=rmse_test_loss_per_question)
