import torch
from sklearn.model_selection import KFold
from sklearn.preprocessing import MinMaxScaler
from torch.nn.utils.rnn import pad_sequence
from RNNUtils import Constants, SimpleRNN
from GeneralUtils import create_tensor_from, plot_loss_per_question, scale_features, get_dev_data, \
    DatasetDirectoryNames, get_dev_training_data
import statistics as st
import matplotlib.pyplot as plt

K_FOLDS = 12
K_FOLD_SEED = 1
VAL_RATIO = 0.3
should_scale_features = True
useTunedHyperparameters = True

hyperparam_directory = DatasetDirectoryNames.firstLastNansInterpolatedTrend5
tunedHyperparamsFile = f"best_params_{hyperparam_directory}.csv"

if __name__ == "__main__":
    torch.manual_seed(Constants.torch_seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    if useTunedHyperparameters:
        print(f"Using AutoTuned Hyperparameters from {tunedHyperparamsFile}.")
        Constants.apply_best_param_config(tunedHyperparamsFile)

    else:
        print(f"Not using autotuned hyperparameters.")

    dev_data_folder = DatasetDirectoryNames.firstLastNansInterpolatedTrend5
    X_train_df, y_train_df = get_dev_training_data(dev_data_folder)

    patient_ids = X_train_df['patient_id'].unique()
    kf = KFold(n_splits=K_FOLDS, random_state=K_FOLD_SEED, shuffle=True)

    epochsScores = {}

    for n_epochs in range(Constants.n_epochs + 5, Constants.n_epochs + 10, 1):
        print("Trial for n_epochs: ", n_epochs)
        cross_val_scores = []
        for train_index, validation_index in kf.split(patient_ids):
            train_ids, val_ids = patient_ids[train_index], patient_ids[validation_index]

            X_train = X_train_df[X_train_df['patient_id'].isin(train_ids)]
            y_train = y_train_df[y_train_df['patient_id'].isin(train_ids)]

            X_val = X_train_df[X_train_df['patient_id'].isin(val_ids)].groupby('patient_id').first().reset_index()
            y_val = y_train_df[y_train_df['patient_id'].isin(val_ids)].groupby('patient_id').first().reset_index()

            print("Number of train patients:", len(train_ids))
            print("Number of val patients:", len(val_ids))

            print("Entries(rows) in train: ", len(X_train))
            print("Entries(rows) in val: ", len(X_val))

            if should_scale_features:
                # scaler = MinMaxScaler(feature_range=(-1, 1))
                X_train, X_val = scale_features(X_train, X_val)

            # Convert data to lists of torch tensors (prepare for training)
            X_train_tensors = [create_tensor_from(group) for _, group in X_train.groupby('patient_id')]
            y_train_tensors = [create_tensor_from(group) for _, group in y_train.groupby('patient_id')]
            X_test_tensors = [create_tensor_from(group) for _, group in X_val.groupby('patient_id')]
            y_test_tensors = [create_tensor_from(group) for _, group in y_val.groupby('patient_id')]

            # Pad the sequences
            X_train = pad_sequence(X_train_tensors, batch_first=True)
            y_train = pad_sequence(y_train_tensors, batch_first=True)

            X_test = pad_sequence(X_test_tensors, batch_first=True)
            y_test = pad_sequence(y_test_tensors, batch_first=True)

            n_features = X_train.shape[-1]
            prediction_length = y_train.shape[-1]

            model = SimpleRNN(n_layers=Constants.n_layers,
                              hidden_size=Constants.hidden_size,
                              input_size=n_features,
                              prediction_length=prediction_length,
                              dropout=Constants.dropout)

            optimizer = torch.optim.Adam(model.parameters(), lr=Constants.learning_rate)

            # Training
            for epoch in range(n_epochs):
                model.zero_grad()
                output = model(X_train)

                loss = model.mean_mae_rmse_loss(output, y_train)
                loss.backward()

                # Gradient clipping
                if Constants.should_clip_gradient:
                    torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1)

                optimizer.step()
                # if epoch % 10 == 0:
                #     print(f'Epoch: {epoch}, Avg(MAE, RMSE) Loss: {loss.item()}')

            # Evaluation
            model.eval()
            with torch.no_grad():
                predictions = model(X_test)
                predictions = torch.round(predictions)
                predictions = torch.clamp(predictions,
                                          min=Constants.min_prediction,
                                          max=Constants.max_prediction)

                mae_test_loss = model.mae_loss(predictions, y_test)
                rmse_test_loss = model.rmse_loss(predictions, y_test)

                print(f'\nTest Loss Avg(MAE, RMSE): {model.mean_mae_rmse_loss(predictions, y_test)}\n')

                print(f'\nMAE Test Loss: {mae_test_loss.item()}')
                print(f'RMSE Test Loss: {rmse_test_loss.item()} \n')
                cross_val_scores.append(rmse_test_loss.item())

        cross_val_score = st.mean(cross_val_scores)
        epochsScores[n_epochs] = cross_val_score


    # print("MEAN Test Loss over epoch trials: ", st.mean(rmse_test_losses.values()))

    epochs = list(epochsScores.keys())
    losses = list(epochsScores.values())

    plt.figure(figsize=(10, 6))
    plt.plot(epochs, losses, marker='o')
    plt.title('Validation Loss over Epoch Trials')
    plt.xlabel('Number of Epochs')
    plt.ylabel('RMSE Validation Loss')
    plt.xticks(epochs)
    plt.grid(True)
    plt.show()

