import torch
from torch.nn.utils.rnn import pad_sequence
from RNNUtils import Constants, SimpleRNN
from GeneralUtils import create_tensor_from, plot_loss_per_question, scale_features, get_dev_data, DatasetDirectoryNames

should_scale_features = True
useTunedHyperparameters = False

dev_data_folder = DatasetDirectoryNames.interpolatedSPODroppedTrend5
tunedHyperparamsFile = f"best_params_{dev_data_folder}.csv"

if __name__ == "__main__":
    torch.manual_seed(Constants.torch_seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    X_train, y_train, X_test, y_test = get_dev_data(dev_data_folder)

    if should_scale_features:
        # scaler = MinMaxScaler(feature_range=(-1, 1))
        X_train, X_test = scale_features(X_train,
                                         X_test)

    if useTunedHyperparameters:
        print(f"Using AutoTuned Hyperparameters from {tunedHyperparamsFile}.")
        Constants.apply_best_param_config(tunedHyperparamsFile)
    else:
        print(f"Not using autotuned hyperparameters.")

    # Convert data to lists of torch tensors (prepare for training)
    X_train_tensors = [create_tensor_from(group) for _, group in X_train.groupby('patient_id')]
    y_train_tensors = [create_tensor_from(group) for _, group in y_train.groupby('patient_id')]
    X_test_tensors = [create_tensor_from(group) for _, group in X_test.groupby('patient_id')]
    y_test_tensors = [create_tensor_from(group) for _, group in y_test.groupby('patient_id')]

    # Pad the sequences
    X_train = pad_sequence(X_train_tensors, batch_first=True)
    y_train = pad_sequence(y_train_tensors, batch_first=True)

    X_test = pad_sequence(X_test_tensors, batch_first=True)
    y_test = pad_sequence(y_test_tensors, batch_first=True)

    n_features = X_train.shape[-1]
    prediction_length = y_train.shape[-1]

    model = SimpleRNN(n_layers=Constants.n_layers,
                      hidden_size=Constants.hidden_size,
                      input_size=n_features,
                      prediction_length=prediction_length,
                      dropout=Constants.dropout)

    optimizer = torch.optim.Adam(model.parameters(), lr=Constants.learning_rate)

    # Training
    for epoch in range(Constants.n_epochs):
        model.zero_grad()
        output = model(X_train)

        loss = model.mean_mae_rmse_loss(output, y_train)
        loss.backward()

        # Gradient clipping
        if Constants.should_clip_gradient:
            torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1)

        optimizer.step()
        if epoch % 10 == 0:
            print(f'Epoch: {epoch}, Avg(MAE, RMSE) Loss: {loss.item()}')

    # Evaluation
    model.eval()
    with torch.no_grad():
        predictions = model(X_test)
        predictions = torch.round(predictions)
        predictions = torch.clamp(predictions,
                                  min=Constants.min_prediction,
                                  max=Constants.max_prediction)

        mae_test_loss = model.mae_loss(predictions, y_test)
        rmse_test_loss = model.rmse_loss(predictions, y_test)

        print(f'\nTest Loss Avg(MAE, RMSE): {model.mean_mae_rmse_loss(predictions, y_test)}\n')

        print(f'\nMAE Test Loss: {mae_test_loss.item()}')
        print(f'RMSE Test Loss: {rmse_test_loss.item()} \n')

        # Initialize lists to store MAE and RMSE for each question
        mae_test_loss_per_question = []
        rmse_test_loss_per_question = []
        total_loss_per_question = []

        # Loop over the third dimension of the tensors (i.e., the questions)
        print(predictions.shape)
        for i in range(predictions.shape[2]):
            mae_test_loss = model.mae_loss(predictions[:, :, i], y_test[:, :, i])
            rmse_test_loss = model.rmse_loss(predictions[:, :, i], y_test[:, :, i])
            total_loss = (mae_test_loss.item() + rmse_test_loss.item()) / 2

            mae_test_loss_per_question.append(mae_test_loss.item())
            rmse_test_loss_per_question.append(rmse_test_loss.item())
            total_loss_per_question.append(total_loss)

        plot_loss_per_question(total_losses=total_loss_per_question,
                               mae_losses=mae_test_loss_per_question,
                               rmse_losses=rmse_test_loss_per_question,
                               should_print=False)
