import optuna
import torch
from torch.nn.utils.rnn import pad_sequence
import pandas as pd
from sklearn.model_selection import KFold
import statistics as st

from RNNUtils import Constants, SimpleRNN
from GeneralUtils import DatasetDirectoryNames, get_dev_training_data, scale_features, \
    create_tensor_from
from functools import partial

data_directory = DatasetDirectoryNames.firstLastNansInterpolatedTrend5
should_normalize_features = True
kfold_seed = 1  # The kfold seed is only useful if kfold_shuffle is True
kfold_shuffle = True


def objective(trial, k_folds):
    print(f"Starting trial: {trial.number}".format(trial))
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.manual_seed(Constants.torch_seed)

    should_clip_gradient = trial.suggest_categorical('clip_gradient', [True, False])
    trial_lr = trial.suggest_float('learning_rate', 0.01, 0.1, step=0.01)
    trial_hidden_size = trial.suggest_int('hidden_size', 2, 30, step=2)
    trial_n_epochs = trial.suggest_int('n_epochs', 1, 301, step=10)
    trial_n_layers = trial.suggest_int('n_layers', 1, 5)
    trial_dropout = trial.suggest_float('dropout', 0.1, 0.5, step=0.05)

    if trial_n_layers <= 1:
        trial_dropout = 0.0  # dropout does not apply to last layer anyway

    X_train_df, y_train_df = get_dev_training_data(data_directory)

    patient_ids = X_train_df['patient_id'].unique()

    # cross validation
    kf = KFold(n_splits=k_folds, random_state=kfold_seed, shuffle=kfold_shuffle)
    cross_validation_losses = []

    for train_index, validation_index in kf.split(patient_ids):
        train_ids, validation_ids = patient_ids[train_index], patient_ids[validation_index]

        # Create train and validation dataframes based on the split patient_ids
        X_train = X_train_df[X_train_df['patient_id'].isin(train_ids)]
        y_train = y_train_df[y_train_df['patient_id'].isin(train_ids)]
        X_val = X_train_df[X_train_df['patient_id'].isin(validation_ids)].groupby('patient_id').first().reset_index()
        y_val = y_train_df[y_train_df['patient_id'].isin(validation_ids)].groupby('patient_id').first().reset_index()

        if should_normalize_features:
            # minmax = MinMaxScaler(feature_range=(-1, 1))
            X_train, X_val = scale_features(X_train, X_val)

        X_train_tensors = [create_tensor_from(group) for _, group in X_train.groupby('patient_id')]
        y_train_tensors = [create_tensor_from(group) for _, group in y_train.groupby('patient_id')]
        X_val_tensors = [create_tensor_from(group) for _, group in X_val.groupby('patient_id')]
        y_val_tensors = [create_tensor_from(group) for _, group in y_val.groupby('patient_id')]

        # Pad the sequences
        X_train = pad_sequence(X_train_tensors, batch_first=True)
        y_train = pad_sequence(y_train_tensors, batch_first=True)

        X_val = pad_sequence(X_val_tensors, batch_first=True)
        y_val = pad_sequence(y_val_tensors, batch_first=True)

        n_features = X_train.shape[-1]
        prediction_length = y_train.shape[-1]

        model = SimpleRNN(n_layers=trial_n_layers,
                          hidden_size=trial_hidden_size,
                          input_size=n_features,
                          prediction_length=prediction_length,
                          dropout=trial_dropout)

        optimizer = torch.optim.Adam(model.parameters(), lr=trial_lr)

        # Training
        for epoch in range(trial_n_epochs):
            model.zero_grad()
            output = model(X_train)

            loss = model.mean_mae_rmse_loss(output, y_train)
            loss.backward()

            if should_clip_gradient:
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1)

            optimizer.step()

        # Validation
        model.eval()
        with torch.no_grad():
            predictions = model(X_val)
            predictions = torch.round(predictions)
            predictions = torch.clamp(predictions,
                                      min=Constants.min_prediction,
                                      max=Constants.max_prediction)

            validation_loss = model.mean_mae_rmse_loss(predictions, y_val)
            cross_validation_losses.append(validation_loss.item())

    mean_val_loss = st.mean(cross_validation_losses)
    print(f'Trial {trial.number} Mean Validation Loss: {mean_val_loss} \n')

    return mean_val_loss


if __name__ == "__main__":
    study = optuna.create_study(direction='minimize')
    try:
        study.optimize(partial(objective, k_folds=12), n_trials=6000)
    except KeyboardInterrupt:
        print("\nOptimization interrupted by user.")
    finally:
        print("Study statistics")
        trial = study.best_trial
        best_params = study.best_params

        df = pd.DataFrame.from_dict(best_params, orient='index').transpose()
        df.to_csv(f'best_params_{data_directory}.csv', index=False)

        print("Best trial mean cross-val MAE: ", trial.value)
        print("Best trial params: ")

        for param in best_params:
            print(f"{param}: {best_params[param]}")

            # Save data from all trials to a CSV file
            df_all_trials = study.trials_dataframe()
            df_all_trials.to_csv(f'all_trials_{data_directory}.csv', index=False)