#%%
import pandas as pd
from sklearn.model_selection import train_test_split
import torch
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler


class DatasetDirectoryNames:
    firstLastNans = "first_last_nans"
    firstLastNansDropped = "first_last_nans_dropped"
    interpolatedSPODropped = "first_last_nans_interpolated_spo_dropped"
    firstLastNansInterpolatedTrend5 = "first_last_nans_interpolated_trend5_strip87"
    firstLastNansInterpolatedTrend3 = "first_last_nans_interpolated_trend3_strip87"
    allNonconsecFirstLastNans = "nonconsec_first_last_nans_interpolated_spo_dropped_strip87"


def get_dev_data(dataset_directory):
    X_train = pd.read_csv(f'../../dev_data/{dataset_directory}/X_train.csv')
    y_train = pd.read_csv(f'../../dev_data/{dataset_directory}/Y_train.csv')
    X_test = pd.read_csv(f'../../dev_data/{dataset_directory}/X_test.csv')
    y_test = pd.read_csv(f'../../dev_data/{dataset_directory}/Y_test.csv')

    return X_train, y_train, X_test, y_test


def get_dev_training_data(dataset_directory):
    X_train = pd.read_csv(f'../../dev_data/{dataset_directory}/X_train.csv')
    y_train = pd.read_csv(f'../../dev_data/{dataset_directory}/Y_train.csv')

    return X_train, y_train


def drop_columns(df, first=False, last=False, trend=False):
    df_res = df.copy()
    if first:
        df_res = df_res.drop(
            [i for i in list(df_res.columns) if i.startswith('init_') and i not in ['init_t', 'init_q']],
            axis='columns')
    if last:
        df_res = df_res.drop(
            [i for i in list(df_res.columns) if i.startswith('target_') and i not in ['target_t', 'target_q']],
            axis='columns')
    if trend:
        df_res = df_res.drop([i for i in list(df_res.columns) if 'poly_trend' in i], axis='columns')
    return df_res


def drop_high_zero_null_columns_greater_than_threshold(df, threshold):
    # Calculate percentage of zero or null values for each column
    percentage_zero_null = ((df == 0) | df.isnull()).mean() * 100

    # Identify columns to drop (percentage > threshold)
    columns_to_drop = percentage_zero_null[percentage_zero_null > threshold].index

    # Print dropped column names
    print(f"Dropped columns: {', '.join(columns_to_drop)}")

    # Drop identified columns and return the resulting DataFrame
    return df.drop(columns_to_drop, axis=1)


def scale_features(train_sets, test_sets, scaler=StandardScaler()):
    scaled_train_sets = []
    scaled_test_sets = []

    for X_tr, X_te in zip(train_sets, test_sets):
        X_train = X_tr.copy()
        X_test = X_te.copy()

        # Assumes patient_id is the only non-numerical column
        numerical_cols = [col for col in X_train.columns if col != 'patient_id']

        X_train_numerical = scaler.fit_transform(X_train[numerical_cols])
        X_test_numerical = scaler.transform(X_test[numerical_cols])

        X_train[numerical_cols] = X_train_numerical
        X_test[numerical_cols] = X_test_numerical

        scaled_train_sets.append(X_train)
        scaled_test_sets.append(X_test)

    return scaled_train_sets, scaled_test_sets



def plot_loss_per_question(total_losses, mae_losses=None, rmse_losses=None, should_print=True):
    if mae_losses is None:
        mae_losses = []
    if rmse_losses is None:
        rmse_losses = []
    mae_losses_count = len(mae_losses)
    rmse_losses_count = len(rmse_losses)

    if should_print:
        print("Test stats per question")
        for i in range(mae_losses_count):
            print(f'\nQuestion {i + 1} MAE Test Loss: {mae_losses[i]}')
            if i < rmse_losses_count:
                print(f'Question {i + 1} RMSE Test Loss: {rmse_losses[i]}')

    total_losses_count = len(total_losses)

    plt.figure(figsize=(10, 6))
    plt.bar(range(1, total_losses_count + 1), total_losses, color='blue', alpha=0.7)
    plt.xlabel('Question Number')
    plt.ylabel('Test Loss')
    plt.title('Avg Test Loss (MAE+RMSE) for each Question')
    plt.xticks(range(1, 13))

    plt.savefig('Test_Loss_Per_Question.png')
    plt.show()


def create_tensor_from(patient_df_group):
    # Drop patient_id column
    df = patient_df_group.drop(['patient_id'], axis=1)

    # Convert DataFrame to tensor
    tensor_df = torch.tensor(df.values, dtype=torch.float32)
    return tensor_df


def fillNansBasic(df):
    if df.isna().any().any():
        print("There are NaN values in the DataFrame. Applying forward fill.")
        df.ffill(inplace=True)
        if df.isna().any().any():
            print("There are still NaN values in the DataFrame after forward fill. Applying backwards fill.")
            df.bfill(inplace=True)
        else:
            print("There are no NaN values in the DataFrame after forward filling.")
    else:
        print("There are no NaN values in the DataFrame.")

    return df


# generates a sub file
# NOTE: uses patient_id column from X_test

def generate_submission(X_test, pred, config):
    predictions = pred.copy()
    qq = predictions.columns.tolist()
    desc = 'mandatory_T1_' + config
    predictions['patient_id'] = X_test['patient_id'].copy()
    predictions['desc'] = desc
    cols = ['patient_id'] + qq + ['desc']
    predictions = predictions[cols]
    print(predictions)
    predictions.to_csv(f"../potential_subs/{desc}.txt", sep=' ', index=False, header=False)


def stitch_dfs(X_train, X_test, y_train, y_test):
    y_combined = pd.concat([y_train, y_test], axis=0)
    X_combined = pd.concat([X_train, X_test], axis=0)
    return X_combined, y_combined
